#!/usr/bin/perl
use strict;
use XButtler::Common;
use XButtler::LoadConfig;
use XButtler::GtkMain;
use XButtler::Log;
use Data::Dumper;
use Gtk;
use Gtk::Gdk;
use POE;
use POE::Component::IRC;
use Getopt::Std;
use vars qw($opt_d $opt_c $opt_h $opt_u $opt_U $opt_m $opt_M);


use constant VERSION				=> "0.82";
use constant AUTHOR_EMAIL			=> "ackmed\@gotwalls.com";

use constant DEFAULT_IRCPORT			=> 6667;

use constant NETWORK_MAX_ATTEMPT		=>  5;   # give up after 5 times trying to connect to a network

use constant NETWORK_STATUS_CONNECTED		=>  1;
use constant NETWORK_STATUS_DISCONNECTED	=>  2;
use constant NETWORK_STATUS_DISABLED		=>  3;
use constant NETWORK_STATUS_CONNECTING		=>  4;

use constant CHANNEL_STATUS_SUCCESS		=>  1;
use constant CHANNEL_STATUS_JOIN		=>  2;
use constant CHANNEL_STATUS_BANNED		=>  3;
use constant CHANNEL_STATUS_INVITE		=>  4;
use constant CHANNEL_STATUS_TOOMANY		=>  5;
use constant CHANNEL_STATUS_FULL		=>  6;
use constant CHANNEL_STATUS_ERROR		=>  7;
use constant CHANNEL_STATUS_KICKED		=>  8;

use constant NICK_LENGTH_MIN			=>  6;
use constant NICK_LENGTH_MAX			=> 10;
use constant USERNAME_LENGTH_MIN		=>  6;
use constant USERNAME_LENGTH_MAX		=> 10;

use constant QUEUE_STATE_ERROR			=> -1;
use constant QUEUE_STATE_NEW			=>  0;	# new entry in the queue
use constant QUEUE_STATE_REMOTE			=>  1;	# queue'd on the bot
use constant QUEUE_STATE_LOCAL			=>  2;	# queue'd locally
use constant QUEUE_STATE_REQUEST		=>  3;	# sent request to bot
use constant QUEUE_STATE_FULL			=>  4;	# bot's queue is full
use constant QUEUE_STATE_TRANSFER		=>  5;	# transfering the file
use constant QUEUE_STATE_OFFLINE		=>  6;	# bot is offline
use constant QUEUE_STATE_REMOVE			=>  7;	# remove queue item
use constant QUEUE_STATE_TIMEOUT		=>  8;  # timeout on the request to bot

use constant QUEUE_RUN_INTERVAL			=> 1;	 # how often to run the queue
use constant QUEUE_RETRY_INTERVAL		=> 1200; # 20mins, how often to re-request a pack from a bot
use constant QUEUE_REQUEST_TIMEOUT		=> 300;  # 5mins, how long the bot is allowed to respond to a request
use constant QUEUE_ERROR_TIMEOUT		=> 300;  # 5mins, how long to keep ERROR'd queue items around

use constant QUEUE_ALIAS			=> "queue";

use constant BOT_STATUS_ONLINE			=> 1;
use constant BOT_STATUS_OFFLINE			=> 2;

use constant GTK_OFFER_REFRESH			=> 30;   # how often to refresh offer clist
use constant GTK_OFFER_STALE_TIME		=> 1200; # 20 mins
use constant GTK_OFFER_COLOR_NETWORK_OFFLINE	=> "#FF0000";
use constant GTK_OFFER_COLOR_BOT_OFFLINE	=> "#FF0000";
use constant GTK_OFFER_COLOR_STALE_OFFER	=> "#FFFF00";
use constant GTK_OFFER_COLOR_DEFAULT		=> "#FFFFFF";
use constant GTK_OFFER_COLOR_QUEUE		=> "#00FF00";

use constant GTK_QUEUE_STALE_TIME		=> 1200; # 20mins
use constant GTK_QUEUE_COLOR_STALE_REQUEST	=> "#FFFF00";
use constant GTK_QUEUE_COLOR_BOT_OFFLINE	=> "#FF0000";
use constant GTK_QUEUE_COLOR_NETWORK_OFFLINE	=> "#FF0000";
use constant GTK_QUEUE_COLOR_DEFAULT		=> "#FFFFFF";


use constant GTK_ALIAS				=> "gtkmain";


# TODO
# - better gui
#	+ 4 tabs, Offers, queue, transfers, networks
#	+ network debug window?
#	+ multiselect from clist, right click brings up menu
# - when gui done, dont have it auto connect to networks on startup
# - make system for handling incoming dcc files
#	+ deal with dupes
#	+ deal with incompletes
# - make queue_dump output better
# - resolve the unknown offer messages
# - how to deal with offers when disconnected from server
#	+ remove offers?
# - how to deal with bot quiting network
#	+ remove offers?
# - add proactive xdcc list requests?
# - ability to try and rejoin channels
# - DCC resume support
# - ability to cancel a request 
# - config work
#	+ config editor
# - only accept DCC from people in the queue
# - add option to make voice something to match on for bot filter
# - better way to track bots in channels
# - disable network/server is unable to connect after a givin amount of time
# - disable users if we any sometype of error with them in a previous dcc session
# - try and get cpu usage down.  maybe nfs related for me, can only write to files in 512byte blocks
# - ?

# global vars
my $config;
my $dcc;
my $stats;
my $debug;
# unknown message
my $unknown;
# msgs from users in the queue
my $mqueue;


my $butt;

$|=1;
main();

sub main {
	print "XButtler Version " . VERSION . " by " . AUTHOR_EMAIL . "\n";

	getopts('c:dhu:U:m:M:');

	if(defined($opt_h)) {
		usage();
		exit;	
	}

	if(defined($opt_d)) {
		$debug = TRUE;
		$debug && print trace(), "Debug mode activated\n";
	}

	if(!defined($opt_c)) {
		usage();
		exit;
	}

	if(defined($opt_u) && defined($opt_U)) {
		print "Error:  Cannot use both option -u and -U at the same time.\n";
		exit;
	} elsif(defined($opt_u)) {
		$unknown = XButtler::Log->new(File => $opt_u);
		if(!defined($unknown)) {
			print "Error: Unable to open $opt_u for writing.\n";
			exit;
		}
	} elsif(defined($opt_U)) {
		$unknown = XButtler::Log->new(File => $opt_U, Append => TRUE);
		if(!defined($unknown)) {
			print "Error: Unable to open $opt_U for writing.\n";
			exit;
		}
	}		

	if(defined($opt_m) && defined($opt_M)) {
		print "Error:  Cannot use both option -m and -M at the same time.\n";
		exit;
	} elsif(defined($opt_m)) {
		$mqueue = XButtler::Log->new(File => $opt_m);
		if(!defined($mqueue)) {
			print "Error: Unable to open $opt_m for writing.\n";
			exit;
		}
	} elsif(defined($opt_M)) {
		$mqueue = XButtler::Log->new(File => $opt_M, Append => TRUE);
		if(!defined($mqueue)) {
			print "Error: Unable to open $opt_M for writing.\n";
			exit;
		}
	}

	$config = LoadConfig(ConfigDir => $opt_c, Debug => $debug);

#	print Dumper($config);
#	exit;

	
	if(!defined($config)) {
		print "Error Loading '$config' xml file.\n";
		exit;
	}

	my $gtk_session = POE::Session->create (
		inline_states => {
			_start			=> \&gtk_start,
			gtk_channel_refresh	=> \&gtk_channel_refresh,
			gtk_dcc_refresh		=> \&gtk_dcc_refresh,
			gtk_network_refresh	=> \&gtk_network_refresh,
			gtk_queue_refresh	=> \&gtk_queue_refresh,
			gtk_statusbar_refresh	=> \&gtk_statusbar_refresh,

			gtk_offer_refresh	=> \&gtk_offer_refresh,
			gtk_offer_row_add	=> \&gtk_offer_row_add,

			gtk_offer_popup 	=> \&gtk_offer_popup,
			gtk_queue_popup		=> \&gtk_queue_popup,
			gtk_dcc_popup		=> \&gtk_dcc_popup,
			gtk_network_popup	=> \&gtk_network_popup,
			gtk_channel_popup	=> \&gtk_channel_popup,

			gtk_row_leftclick	=> \&gtk_row_leftclick,
			gtk_row_rightclick	=> \&gtk_row_rightclick,
			gtk_row_select		=> \&gtk_row_select,
			gtk_row_unselect	=> \&gtk_row_unselect,

			gtk_search_new		=> \&gtk_search_new,
			gtk_status		=> \&gtk_status,
			gtk_title_leftclick	=> \&gtk_title_leftclick,
		},);
			
	$stats->{'TOTAL_PACKS'} = 0;
	$stats->{'TOTAL_NETWORKS'} = 0;
	$stats->{'TOTAL_CHANNELS'} = 0;
	$stats->{'TOTAL_DOWNLOAD'} = 0;
	$stats->{'SPEED'} = 0.0;

	# have config, create poe sessions
	for my $network (keys %{$config->{'NETWORKS'}}) {
		POE::Component::IRC->new("$network-bot");
		POE::Session->create(
			inline_states => {
				_start 			=> \&network_start,
				irc_001 		=> \&network_connected,
				irc_332			=> \&channel_join_topic,
				irc_353			=> \&channel_join_list,
				irc_366			=> \&channel_join_success,
				irc_401			=> \&user_offline,
				irc_403			=> \&channel_join_error,
				irc_405			=> \&channel_too_many,
				irc_461			=> \&channel_join_error,
				irc_471			=> \&channel_full,
				irc_473			=> \&channel_invite_only,
				irc_474			=> \&channel_banned,
				irc_475			=> \&channel_join_error,
				irc_476			=> \&channel_join_error,
				irc_ctcp_version	=> \&ctcp_version,
				irc_dcc_done		=> \&dcc_done,
				irc_dcc_error		=> \&dcc_error,
				irc_dcc_get		=> \&dcc_get,
				irc_dcc_request		=> \&dcc_request,
				irc_dcc_start		=> \&dcc_start,
				irc_disconnected	=> \&network_disconnected,
				irc_error		=> \&network_error,
				irc_join		=> \&channel_join,
				irc_msg			=> \&network_parsemsg,
				irc_notice		=> \&network_parsemsg,
				irc_part		=> \&channel_part,
				irc_public		=> \&network_parsemsg,
				irc_quit		=> \&channel_quit,
				irc_socketerr		=> \&network_socket_error,
				network_connect		=> \&network_connect,
			},
			args => [ $network ],
		);
	}

	# create queue session
	POE::Session->create(
		inline_states => {
			_start			=> \&queue_start,
			queue_add_item		=> \&queue_add_item,
			queue_clear		=> \&queue_clear,
			queue_dump		=> \&queue_dump,
			queue_get_queue		=> \&queue_get_queue,
			queue_incoming_dcc	=> \&queue_incoming_dcc,
			queue_valid_user	=> \&queue_valid_user,
			queue_remove_item	=> \&queue_remove_item,
			queue_run		=> \&queue_run,
			queue_update_item	=> \&queue_update_item,
		},);
	
	# start the fun
	$poe_kernel->run();
}

##
## irc network connection stuff
##

# setup network, and issue network_connect signal
sub network_start {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $session = @_[SESSION];
	my $network = @_[ARG0];
	
	# save network on our session heap
	$heap->{'NETWORK'} = $network;

	$kernel->alias_set("$network-session");
	$kernel->post("$network-bot" => register => "all");

	$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISCONNECTED;
	$butt->{'NETWORKS'}->{$network}->{'INFO'} = "";
	$butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} = 0;

	# start the connection process
	$kernel->delay("network_connect",rand(2 * scalar(keys %{$config->{'NETWORKS'}})));
	$kernel->delay("network_blah",30);
}

# setups up the network connection
sub network_connect {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $network = $heap->{'NETWORK'};


	$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISCONNECTED;

	# start connection process to server
	my $server = random_element(@{$config->{'NETWORKS'}->{$network}->{'SERVERS'}});
	
	my ($server_ip,$server_port,$server_pass) = split(":",$server);

	if(!defined($server_port)) {
		$server_port = DEFAULT_IRCPORT;
	}
	
	my $nick = $config->{'NETWORKS'}->{$network}->{'NICK'};
	if($nick eq "RANDOM" || !defined($nick)) {
		$nick = random_string(NICK_LENGTH_MIN,NICK_LENGTH_MAX);
	}
	
	my $username = $config->{'NETWORKS'}->{$network}->{'USERNAME'};
	if($username eq "RANDOM") {
		$username = random_string(USERNAME_LENGTH_MIN,USERNAME_LENGTH_MAX);
	} elsif($username eq "NICK") {
		$username = $nick;
	}

	print "$network: Connecting to $server as $nick\n";
	$kernel->post(
		"$network-bot" => connect => {
			Nick		=> $nick,
			Server		=> $server_ip,
			Port		=> $server_port,
			Password	=> $server_pass,
			Ircname		=> $config->{'NETWORKS'}->{$network}->{'IRCNAME'},,
			Username	=> $username,
		}
	);

	# save nick
	$heap->{'NICK'} = $nick;
	$butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'}++;
}

# called when we are fully connect to the irc server
sub network_connected {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $network = $heap->{'NETWORK'};

	# join channels
	my @channels = keys(%{$config->{'NETWORKS'}->{$network}->{'CHANNELS'}});
	print "$network: Connected. Joining (@channels)\n";

	$kernel->post("$network-bot" => mode => "$heap->{'NICK'} $config->{'NETWORKS'}->{$network}->{'UMODE'}");

	# stats
	$stats->{'TOTAL_NETWORKS'}++;
	$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_CONNECTED;
	$butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} = 0;
	$butt->{'NETWORKS'}->{$network}->{'INFO'} = "";

	$kernel->post(GTK_ALIAS,"gtk_status","Network: $network connection successful.");

	for my $channel (@channels) {
		# add channel key if needed
		if(defined($config->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'KEY'})) {
			$channel = $channel . " " . $config->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'KEY'};
		}
		$kernel->post("$network-bot" => join => $channel);
		$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_JOIN;
		$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'TOPIC'} = "";
	}
}

# called when we are disconnected from irc server
sub network_disconnected {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $server = @_[ARG0];
	my $reason = @_[ARG1];
	my $network = $heap->{'NETWORK'};
	

	# someone else already started the reconnect process
	if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_CONNECTING || $butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISABLED) {
		return;
	}

	# stats
	if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_CONNECTED) {
		$stats->{'TOTAL_NETWORKS'}--;
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISCONNECTED;
		$kernel->post(GTK_ALIAS,"gtk_status","Network: $network Disconnected ($reason)");
		for my $channel (keys %{$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}}) {
			if($butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} == CHANNEL_STATUS_SUCCESS) {
				$stats->{'TOTAL_CHANNELS'}--;
			}
			
		}
		delete($butt->{'NETWORKS'}->{$network}->{'CHANNELS'});
	}

	# save reason/error 
	if($reason ne "") {
		$butt->{'NETWORKS'}->{$network}->{'INFO'} = $reason;
	}

	print "$network: Disconnect from $server ($reason)\n";
	
	if($butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} >= NETWORK_MAX_ATTEMPT) {
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISABLED;
		$kernel->post(GTK_ALIAS,"gtk_status","Network: $network Disabled, unable to connect after $butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} attempts.");
	} elsif($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISCONNECTED) {
		## add scheduler event to reconnect us
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_CONNECTING;
		print "$network: Will Attempt to reconnect in 30 seconds\n";
		$kernel->delay("network_connect",30);
	}
}

# sometype of error occured with the server and we were disconnected
sub network_error {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $error = @_[ARG0];
	my $network = $heap->{'NETWORK'};

	# someone else already started the reconnect process
	if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_CONNECTING || $butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISABLED) {
		return;
	}

	# stats
	if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_CONNECTED) {
		$stats->{'TOTAL_NETWORKS'}--;
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISCONNECTED;
		$kernel->post(GTK_ALIAS,"gtk_status","Network: $network Disconnected, Error ($error)");
		for my $channel (keys %{$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}}) {
			if($butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} == CHANNEL_STATUS_SUCCESS) {
				$stats->{'TOTAL_CHANNELS'}--;
			}
			
		}
		delete($butt->{'NETWORKS'}->{$network}->{'CHANNELS'});
	}

	# save reason/error 
	if($error ne "") {
		$butt->{'NETWORKS'}->{$network}->{'INFO'} = $error;
	}

	print "$network: Disconnect Error ($error)\n";
	
	if($butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} >= NETWORK_MAX_ATTEMPT) {
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISABLED;
		$kernel->post(GTK_ALIAS,"gtk_status","Network: $network Disabled, unable to connect after $butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} attempts.");
	} elsif($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISCONNECTED) {
		## add scheduler event to reconnect us
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_CONNECTING;
		print "$network: Will Attempt to reconnect in 30 seconds\n";
		$kernel->delay("network_connect",30);
	}
}

# sometype of error occured with the server and we were disconnected
sub network_socket_error {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $error = @_[ARG0];
	my $network = $heap->{'NETWORK'};

	# someone else already started the reconnect process
	if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_CONNECTING || $butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISABLED) {
		return;
	}

	# stats
	if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_CONNECTED) {
		$stats->{'TOTAL_NETWORKS'}--;
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISCONNECTED;
		$kernel->post(GTK_ALIAS,"gtk_status","Network: $network Disconnected, Socket Error ($error)");
		for my $channel (keys %{$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}}) {
			if($butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} == CHANNEL_STATUS_SUCCESS) {
				$stats->{'TOTAL_CHANNELS'}--;
			}
			
		}
		delete($butt->{'NETWORKS'}->{$network}->{'CHANNELS'});
	}

	# save reason/error 
	if($error ne "") {
		$butt->{'NETWORKS'}->{$network}->{'INFO'} = $error;
	}

	print "$network: Disconnect Socket Error ($error)\n";
	
	if($butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} >= NETWORK_MAX_ATTEMPT) {
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISABLED;
		$kernel->post(GTK_ALIAS,"gtk_status","Network: $network Disabled, unable to connect after $butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} attempts.");
	} elsif($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISCONNECTED) {
		## add scheduler event to reconnect us
		$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_CONNECTING;
		print "$network: Will Attempt to reconnect in 30 seconds\n";
		$kernel->delay("network_connect",30);
	}

}

##
## channel stuff
##

# successfully joined channel
sub channel_join_success {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};
	
	$data =~ tr/[A-Z]/[a-z]/;
	my $channel = (split(" ",$data))[0];

	$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_SUCCESS;

	# stats
	$stats->{'TOTAL_CHANNELS'}++;
}

sub channel_join_error {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	$data =~ tr/[A-Z]/[a-z]/;
	my $channel = (split(" ",$data))[0];
	if(defined($butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel})) {
		print "Channel ERROR ($channel) : $data\n";
		$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_ERROR;
	}
}

# denied channel join, we are in too many channels
sub channel_too_many {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	$data =~ tr/[A-Z]/[a-z]/;
	my $channel = (split(" ",$data))[0];
	
	$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_TOOMANY;
}

# denied channel join, channel is full
sub channel_full {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	$data =~ tr/[A-Z]/[a-z]/;
	my $channel = (split(" ",$data))[0];
	$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_FULL;
}

# denied channel join, invite only
sub channel_invite_only {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	$data =~ tr/[A-Z]/[a-z]/;
	my $channel = (split(" ",$data))[0];
	$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_INVITE;
}

# denied channel join, banned
sub channel_banned {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	$data =~ tr/[A-Z]/[a-z]/;
	my $channel = (split(" ",$data))[0];
	$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'}  = CHANNEL_STATUS_BANNED;
}

# called when a user joins a channel
sub channel_join {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $user = @_[ARG0];
	my $channel = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	$channel =~ tr/[A-Z]/[a-z]/;
	$user =~ tr/[A-Z]/[a-z]/;

	my ($nick,$host) = split("!",$user);

	## add code to see if the user that just joined was an offline bot
}

# called when we join a channel, gives a list of all users in the channel
sub channel_join_list {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	$data =~ tr/[A-Z]/[a-z]/;

	# $data = '= #channel : user1 user2 ';
	$data =~ s/\://;
	$data =~ s/\=\ //;

	my @nicks = split(/\s+/,$data);
	my $channel = shift @nicks;

}

# called when user gets kicked from channel
sub channel_kick {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $kicker = @_[ARG0];
	my $channel = @_[ARG1];
	my $kickee = @_[ARG2];
	my $reason = @_[ARG3];

	my $network = $heap->{'NETWORK'};
	$kickee =~  tr/[A-Z]/[a-z]/;

	if($kickee eq $heap->{'NICK'}) {
		$debug && print "$network: KICKED from $channel by $kicker ($reason)\n";
		$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_KICKED;
	}

}
# called when user parts a channel
sub channel_part {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $user = @_[ARG0];
	my $channel = @_[ARG1];

	my $network = $heap->{'NETWORK'};

	$channel =~ tr/[A-Z]/[a-z]/;
	my ($nick,$host) = split("!",$user);
	
}

# topic from joining channel
sub channel_join_topic {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];
	my $network = $heap->{'NETWORK'};

	my $channel = (split(" ",$data))[0];
	$channel =~ tr/[A-Z]/[a-z]/;

	# remove colors and stuff
	$data =~ s/\003\d\d,\d\d//g;
	$data =~ s/\003\d,\d\d//g;  
	$data =~ s/\003\d\d,\d//g;
	$data =~ s/\003\d\,\d//g;
	$data =~ s/\003\d\d//g;
	$data =~ s/\003\d//g;
	$data =~ s/\003//g;
	$data =~ s/[\001-\037,\177-\255]//g;

	$data =~ /^.*\ :(.*)/;
	$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'TOPIC'} = $1;
}

# someone quits irc
sub channel_quit {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $user = @_[ARG0];

	my $network = $heap->{'NETWORK'};
	$user =~ tr/[A-Z]/[a-z]/;
        my ($nick,$host) = split("!",$user);

	# mark bot as offline
	$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} = BOT_STATUS_OFFLINE;
}

# parse incoming messages from potential bots
sub network_parsemsg {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $sender = @_[ARG0];
	my $message = @_[ARG2];
	my $network = $heap->{'NETWORK'};

	$sender =~ tr/[A-Z]/[a-z]/;

	my ($nick,$host) = split("!",$sender);

	# remove colors and stuff
	$message =~ s/\003\d\d,\d\d//g;
	$message =~ s/\003\d,\d\d//g;
	$message =~ s/\003\d\d,\d//g;
	$message =~ s/\003\d\,\d//g;
	$message =~ s/\003\d\d//g;
	$message =~ s/\003\d//g;
	$message =~ s/\003//g;
	$message =~ s/[\x00-\x1F]//g;
	$message =~ s/[\x7F-\xFF]//g;
#	$message =~ s/[\001-\037]//g;
#	$message =~ s/[\177-\255]//g;
	

	# '#1  23x [ 66M] release_name'
	# '#1  23x [ <1K] release_name'
	# '#11 66.6MB release_name - (6 gets)'
	# '(#1) 92.8MB  Blondie-grea - (6 gets)'
	# '[#11] (66x|66.6 mb) - release_name'
	if($message =~ /^(\#\d+)\s+\d+x\s+\[\s*([\d\.\<]*.)\]\s+(.*)\s*$/ || $message =~ /^[\(]?(\#\d+)[\)]?\s+([\d\.]*..)\s+(.*)\ -\ \(\d+\s+gets\)/i || $message =~ /\[(\#\d+)\]\s+\(\d+x\|([\d\.]*.*)\)\s+\-\s+(.*)/i) {
		my ($pack,$size,$offer) = ($1,$2,$3);
		# strip off the B from the MB
		$size =~ s/[b\ ]//ig;
		$size =~ tr/[a-z]/[A-Z]/;

		if($pack ne "" && $size ne "" && $offer ne "") {
			my $new_pack = TRUE;
			if(defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack})) {
				$new_pack = FALSE;
			}
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'SIZE'} = $size;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'OFFER'} = $offer;
			

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'} = "?";
			}

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'} = "?";
			}

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'} = "?";
			}

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'} = "?";
			}

			if($new_pack) {
				$kernel->post(GTK_ALIAS,"gtk_offer_row_add",$network,$nick,$pack,$size,$offer);
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'QUEUE'} = FALSE;
				# stats
				$stats->{'TOTAL_PACKS'}++;
			}

			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'LAST_UPDATE'} = time;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} = BOT_STATUS_ONLINE;
		} else {
			$debug && print trace(), "Unable to get valid pack, size, or offer from line '$message'\n";
		}
	# '(#6) - release_name - 66.6MB (6 gets)'
	} elsif($message =~ /\((\#\d+)\)\s+\-\s+(.*)\ \-\ ([\d\.]*.*)\s+\(\d+\ gets\)/i) {
		my ($pack,$size,$offer) = ($1,$3,$2);
		# strip off the B from the MB
		$size =~ s/[b\ ]//ig;
		$size =~ tr/[a-z]/[A-Z]/;

		if($pack ne "" && $size ne "" && $offer ne "") {
			my $new_pack = TRUE;
			if(defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack})) {
				$new_pack = FALSE;
			}
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'SIZE'} = $size;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'OFFER'} = $offer;
			

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'} = "?";
			}

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'} = "?";
			}

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'} = "?";
			}

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'} = "?";
			}

			if($new_pack) {
				$kernel->post(GTK_ALIAS,"gtk_offer_row_add",$network,$nick,$pack,$size,$offer);
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'QUEUE'} = FALSE;
				# stats
				$stats->{'TOTAL_PACKS'}++;
			}

			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'LAST_UPDATE'} = time;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} = BOT_STATUS_ONLINE;
		} else {
			$debug && print trace(), "Unable to get valid pack, size, or offer from line '$message'\n";
		}
	# '** 2 packs **  0 of 5 slots open, Queue: 0/30, Record: 1026.9KB/s'
	} elsif($message =~ /^\*\*\s+(\d+)\s+pack[s]*\s+\*\*\s+(\d+)\sof\s(\d+)\sslot[s]*\sopen\,\s+Queue\:\s+(\d+)\/(\d+)/) {
		my ($total_packs,$open_slots,$total_slots,$used_queue,$total_queue) = ($1,$2,$3,$4,$5);
		if($total_packs ne "" && $open_slots ne "" && $total_slots ne "" && $used_queue ne "" && $total_queue ne "") {
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_PACKS'} = $total_packs;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'} = $open_slots;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'} = $total_slots;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'} = $total_queue - $used_queue;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'} = $total_queue;

			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'LAST_UPDATE'} = time;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} = BOT_STATUS_ONLINE;
		} else {
			$debug && print trace(), "Unable to get valid total packs, total slots, open slots, totel queue, used queue from line '$message'\n";
		}
	# '** 2 packs **  1 of 2 slots open, Record: 113.7KB/s'
        # '** 2 packs **  1 of 1 slot open, Record: 115.1KB/s'
	} elsif($message =~ /^\*\*\s+(\d+)\s+pack[s]*\s+\*\*\s+(\d+)\sof\s(\d+)\sslot[s]*\sopen/) {
		my ($total_packs,$open_slots,$total_slots) = ($1,$2,$3);
		if($total_packs ne "" && $open_slots ne "" && $total_slots ne "") {
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_PACKS'} = $total_packs;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'} = $open_slots;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'} = $total_slots;

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'} = "?";
			}

			if(!defined($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'})) {
				$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'} = "?";
			}

			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'LAST_UPDATE'} = time;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} = BOT_STATUS_ONLINE;
		} else {
			$debug && print trace(), "Unable to get valid total packs, total slots, open slots from line '$message'\n";
		}
	# 'XDCC Server Active Sends:3/3 Queues:15/20'
	# 'sends:[0/1] queues:[6/5] record speed:[23 kb/s by someone] ~(polaris v2.05c)~'
	# '- Sends:(2/2) Queues:(41/50) Record CPS:(43.5Kb/s by someone) ~{Polaris SE}~'
	# 'Sends:1/1 Queues:4/10 Min CPS:9.8 Kbs Record CPS:44.4'
	} elsif($message =~ /XDCC\ Server\ Active\ Sends\:(\d+)\/(\d+)\s+Queues\:(\d+)\/(\d+)/i || $message =~ /sends\:\s*[\[,\(,\s]*(\d+)\/(\d+)[\],.,\)]*.*queues\:[\[,\(,\s]*(\d+)\/(\d+)[\],.,\)]*/i) {
		my ($open_slots,$total_slots,$open_queue,$total_queue) = ($1,$2,$3,$4);
		if($open_slots ne "" && $total_slots ne "" && $open_queue ne "" && $total_queue ne "") {
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'} = $open_slots;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'} = $total_slots;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'} = $open_queue;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'} = $total_queue;

			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'LAST_UPDATE'} = time;
			$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} = BOT_STATUS_ONLINE;

		} else {
			$debug && print trace(), "Unable to get valid total slots, open slots, total queue, open queue from line '$message'\n";
		}
	} elsif($message =~ /\/ctcp\ .*xdcc\ send\ \#/i) {
		# print "$network: $nick wants to use tcp '$message'\n";
	# 'Total Offered: 1351 MB  Total Transferred: 2478 GB'
	} elsif($message =~ /^Total\ Offered/i) {
		# ignore for now
	# '** Bandwidth Usage ** Current: 0.0KB/s, Record: 117.6KB/s
	} elsif($message =~ /\*\*\ Bandwidth\ Usage/i) {
		# ignore for now
	# '** To request a file type: "/msg somenick xdcc send #x" **'
	} elsif($message =~ /\*\*\ To\ request\ a\ file\ type/i) {
		# ignore for now
	# '** . SUPPLIED BY #SOMECHANNEL . **'
	} elsif($message =~ /^\*\*.*SUPPLIED\ BY.*\*\*/i) {
		# ignore
	# '** ** Brought to you by #somechannel ** **'
	} elsif($message =~ /^\*\*.*Brought\ to.*\*\*/i) {
		# ignore

	#
	# bots stuff from here on out, only accept messages from people in the queue
	#
	} elsif($kernel->call(QUEUE_ALIAS,"queue_valid_user",$network,$nick)) {
		# '*** All Slots Full, Added you to the main queue in position 6'
		# 'Queuing release_name (66MB). It has been placed in queue slot #6, it will send when sends are available."
		if($message =~ /^\*\*\*\ All\ Slots\ Full\,\ Added\ you\ to\ the\ main\ queue\ in\ position\s+(\d+)/i || $message =~ /Queuing.*It has been placed in queue slot \#(\d+)/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_REMOTE,"In position $1");

		# 'You have been queued for 6 hr 6 min, currently in main queue position 6 of 6'
		} elsif($message =~ /^You\ have\ been\ queued\ for\s+\d+\s+hr\s+\d+\s+min\,\ currently\ in\ main\ queue\ position\s+(\d+)\ of\s+\d+/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_REMOTE,"In position $1");

		# '*** Transfer Completed (666 kbytes, 6 min 66 sec, 66.6 kbytes/sec)'
		} elsif($message =~ /\*\*\*\ Transfer\ Completed/i) {
	#		$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_REMOVE,"Transfer Complete");

		# '*** Sending You Pack #6 Which Is 66MB. (Resume Supported)'
		# '*** Sending You Your Queued Pack Which Is 66MB. (Resume Supported)'
		# 'DCC Send filename.rar (123.123.123.123)'
		} elsif($message =~ /\*\*\*\ Sending\ You\ Pack\ (\#\d+)/i || $message =~ /^DCC\ Send.*\([\d\.]*\)/i || $message =~ /\*\*\*\ Sending\ You\ Your\ Queued\ Pack/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_TRANSFER,"Transfer Started");

		# '*** All Slots Full, Denied, You already have that item queued.'
		} elsif($message =~ /\*\*\*\ All\ Slots\ Full.*You\ already\ have\ that\ item\ queued/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_ERROR,"Item Already queue'd on bot?");
	
		# '*** All Slots Full, Denied, You already have 1 items queued, Try Again Later'
		} elsif($message =~ /\*\*\*\ All\ Slots\ Full\,\ Denied\,\ You\ already\ have\ \d\ items\ queued\,\ Try\ Again\ Later/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_ERROR,"Already have item queue'd on bot?");
	
		# '*** All Slots Full, Main queue of size 20 is Full, Try Again Later'
		} elsif($message =~ /\*\*\*\ All\ Slots\ Full.*Try\ Again\ Later/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_FULL,"Bot's queue is full");
	
		# '*** Closing Connection: Connection Lost'
		} elsif($message =~ /\*\*\*\ Closing\ Connection\:\ Connection\ Lost/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_ERROR,"Connection Lost: unknown reason");
	
		# '*** Closing Connection: Under Min Speed Requirement, 6.6K/sec is less than 66.6K/sec'
		} elsif($message =~ /\*\*\*\ Closing\ Connection\:\ Under\ Min\ Speed/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_ERROR,"Connection Lost: Under minimum speed requirement");

		# '*** Closing Connection: DCC Timeout (180 Sec Timeout)'
		} elsif($message =~ /\*\*\*\ Closing\ Connection\:\ DCC\ Timeout/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_ERROR,"Connection Lost: DCC Timeout");
	
		# '*** XDCC SEND denied, you must be on a known channel to request a pack'
		} elsif($message =~ /you\ must\ be\ on\ a\ known\ channel/i) {
			$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_ERROR,"XDCC Denied, not in a known channel");
	
		# '*** You have a DCC pending, Set your client to receive the transfer. (150 seconds remaining until timeout)'
		} elsif($message =~ /you\ have\ a\ DCC\ pending/i) {
			$kernel->post(GTK_ALIAS,"gtk_status","DCC: Having problems making connection with $nick\@$network");
		
		# if logging unknown messages of people in the queue
		} elsif(defined($mqueue)) {
			$mqueue->write("[$nick\@$network] $message");
		}
	# log all unknown messages
	} elsif(defined($unknown)) {
		$unknown->write("[$nick\@$network] $message");
	}
}

##
## user stuff
##

# sent a message, but user isnt online?
sub user_offline {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $data = @_[ARG1];

	$data =~ tr/[A-Z]/[a-z]/;

	my $network = $heap->{'NETWORK'};
	my $nick = (split(" ",$data))[0];

	# try to update queue for this user
	$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_OFFLINE,"Bot Offline");
	$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} = BOT_STATUS_OFFLINE;

}

###
### dcc functions
###

# dcc connection finish without error
sub dcc_done {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $dccid = @_[ARG0];
	my $user = @_[ARG1];
	my $dtype = @_[ARG2];
	my $file = @_[ARG4];
	my $size_total = @_[ARG5];

	my $network = $heap->{'NETWORK'};
	my ($nick,$host) = split("!",$user);

	$nick =~ tr/[A-Z]/[a-z]/;

	if($dcc->{'DCC'}->{$dccid}->{'ACTIVE'}) {
		my $timediff = time - $dcc->{'DCC'}->{$dccid}->{'START_TIME'};
		my $speed = sprintf("%.1fK/s",$size_total / ($timediff * 1024));

		# rename file
		my $done = $file;
		$done =~ s/\/transferring\-/\//;

		if(-e $done) {
			$done = $done . "-" . time;
		}

		rename($file,$done);
		
		$file =~ /.*\/transferring\-(.*)/;
		$kernel->post(GTK_ALIAS,"gtk_status","Transfer Complete: $nick - $1 \@ $speed in " . strtimediff($timediff));	
	
		$debug && print trace(), "$nick\@$network $file ($size_total) $speed\n";
		delete($dcc->{'DCC'}->{$dccid});

		$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_REMOVE,"Transfer Complete");
	} elsif(defined($dcc->{'DCC'}->{$dccid})) {
		$nick = $dcc->{'DCC'}->{$dccid}->{'NICK'};
		$nick =~ tr/[A-Z]/[a-z]/;
        	$file = $dcc->{'DCC'}->{$dccid}->{'FILE'};

		# delete file if requested
		if($dcc->{'DCC'}->{$dccid}->{'DELETE'}) {
			unlink($file);
		} else {
			my $stop = $file;
			$stop =~ s/\/transferring\-/\/stopped\-/;
			if(-e $stop) {
				$stop = $stop . "-" . time;
			}
			rename($file,$stop);
		}
	
		$file =~ /.*\/transferring\-(.*)/;
		$kernel->post(GTK_ALIAS,"gtk_status","Transfer Stopped: $nick - $1");

		delete($dcc->{'DCC'}->{$dccid});
		$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_REMOVE,"Transfer Stopped");
	} else {
		print trace(), "DCC: no info on dccid: $dccid\n";
	}
}

# dcc connection error
sub dcc_error {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $dccid = @_[ARG0];
	my $error = @_[ARG1];
	my $user = @_[ARG2];
	my $dtype = @_[ARG3];   
	my $file = @_[ARG5];    
	my $size_total = @_[ARG6];
 
	my $network = $heap->{'NETWORK'};	
	my ($nick,$host) = split("!",$user);
	
	$nick =~ tr/[A-Z]/[a-z]/;

	# rename file
	my $failed = $file;
	$failed =~ s/\/transferring\-/\/failed\-/;

	if(-e $failed) {
		$failed = $failed . "-" . time;
	}

	rename($file,$failed);

	$file =~ /.*\/transferring\-(.*)/;
	$kernel->post(GTK_ALIAS,"gtk_status","Transfer Error: $nick - $1 - $error");
 
	$debug && print trace(), "$nick\@$network $file ($size_total) ($error)\n";
	delete($dcc->{'DCC'}->{$dccid});
	

	# let queue know there was an error
	$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_ERROR,$error);
}

# dcc connection received another block of data
sub dcc_get {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $dccid = @_[ARG0];
	my $user = @_[ARG1];    
	my $file = @_[ARG3];
	my $size_total = @_[ARG4];
	my $size_xfered = @_[ARG5];

	my $network = $heap->{'NETWORK'};
	my ($nick,$host) = split("!",$user);

	$stats->{'TOTAL_DOWNLOAD'} = $stats->{'TOTAL_DOWNLOAD'} + ($size_xfered - $dcc->{'DCC'}->{$dccid}->{'BYTES_READ'});
	$dcc->{'DCC'}->{$dccid}->{'BYTES_READ'} = $size_xfered;

}

# incoming dcc request from another user
sub dcc_request {
	my $kernel = @_[KERNEL];
        my $heap = @_[HEAP];
        my $user = @_[ARG0];
	my $dtype = @_[ARG1];
        my $dccid = @_[ARG3];
        my $file = @_[ARG4];
        my $size_total = @_[ARG5];

	my $network = $heap->{'NETWORK'};
	my ($nick,$host) = split("!",$user);

	$nick =~ tr/[A-Z]/[a-z]/;
	$debug && print trace(), "$nick\@$network $dtype $file ($size_total)\n"; 

	# file already exists.
	if(-e $config->{'NETWORKS'}->{$network}->{'DOWNLOAD_PATH'} . "/transferring-" . $file || -e $config->{'NETWORKS'}->{$network}->{'DOWNLOAD_PATH'} . "/" . $file) {
		$file = time . "-" . $file;
	}

	if($dtype eq "SEND") {
		my $destination = $config->{'NETWORKS'}->{$network}->{'DOWNLOAD_PATH'} . "/transferring-" . $file;
		$kernel->post(QUEUE_ALIAS, "queue_incoming_dcc",$dccid,$nick,$network,$destination);
#		$kernel->post("$network-bot" => dcc_accept => $dccid => $destination);
	}

}

# dcc connection started
sub dcc_start {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP]; 
        my $dccid = @_[ARG0];
        my $user = @_[ARG1]; 
        my $file = @_[ARG4]; 
        my $size_total = @_[ARG5];

	my $network = $heap->{'NETWORK'};
	my ($nick,$host) = split("!",$user);

	$nick =~ tr/[A-Z]/[a-z]/;

	$debug && print trace(), "$nick\@$network $file ($size_total)\n";
	$dcc->{'DCC'}->{$dccid}->{'NICK'} = $nick;
	$dcc->{'DCC'}->{$dccid}->{'FILE'} = $file;
	$dcc->{'DCC'}->{$dccid}->{'BYTES_TOTAL'} = $size_total;
	$dcc->{'DCC'}->{$dccid}->{'BYTES_READ'} = 0;
	$dcc->{'DCC'}->{$dccid}->{'NETWORK'} = $network;
	$dcc->{'DCC'}->{$dccid}->{'START_TIME'} = time();
	$dcc->{'DCC'}->{$dccid}->{'ACTIVE'} = TRUE;
	$dcc->{'DCC'}->{$dccid}->{'SPEED'} = 0.0;

	$file =~ /.*\/transferring\-(.*)/;
	$kernel->post(GTK_ALIAS,"gtk_status","Transfer Starting: $nick - $1");
	$kernel->post(QUEUE_ALIAS,"queue_update_item",$network,$nick,QUEUE_STATE_TRANSFER,"Transfer Started");
}

sub dcc_dump {
#	my $kernel = @_[KERNEL];
#	my $heap = @_[HEAP];

	print "DCC Dump\n-----------------------\n";
	for my $dccid (keys %{$dcc->{'DCC'}}) {
		if(!$dcc->{'DCC'}->{$dccid}->{'ACTIVE'}) {
			next;
		}
		my $timediff = time - $dcc->{'DCC'}->{$dccid}->{'START_TIME'};
		if($timediff <= 0) {
			$timediff = 1;
		}
		my $speed = sprintf("%.1fK/s", $dcc->{'DCC'}->{$dccid}->{'BYTES_READ'} / ($timediff * 1024));
		$dcc->{'DCC'}->{$dccid}->{'FILE'} =~ /.*\/(.*)$/;
		my $file = $1;
		my $percent = sprintf("%.1f%%", 100.0 * $dcc->{'DCC'}->{$dccid}->{'BYTES_READ'} / $dcc->{'DCC'}->{$dccid}->{'BYTES_TOTAL'});
		print "  $dcc->{'DCC'}->{$dccid}->{'NICK'}\@$dcc->{'DCC'}->{$dccid}->{'NETWORK'}: $file @ $speed -  $percent\n";
	}
}

##
## ctcp stuff
##

# ctcp VERSION from another user
sub ctcp_version {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $user = @_[ARG0];
	my $msg = @_[ARG2];

	my $network = $heap->{'NETWORK'};
	my ($nick,$host) = split("!",$user);

	$debug && print trace(), "CTCP VERSION from $nick\@$network\n";
	if(defined($config->{'NETWORKS'}->{$network}->{'VERSION'})) {
		$kernel->post("$network-bot" => ctcpreply => $nick =>  "VERSION " . $config->{'NETWORKS'}->{$network}->{'VERSION'});
	} else { 
		$kernel->post("$network-bot" => ctcpreply => $nick =>  "VERSION Buttler Version " . VERSION);
	}
}

##
## gtk stuff
##

sub gtk_start {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $session = @_[SESSION];

	$kernel->alias_set(GTK_ALIAS);
	$heap->{'GTKOBJ'} = new XButtler::GtkMain();
	$heap->{'WINDOW'} = $heap->{'GTKOBJ'}->get_window();
	$heap->{'OFFER_CLIST'} = $heap->{'GTKOBJ'}->get_offer_clist();
	$heap->{'QUEUE_CLIST'} = $heap->{'GTKOBJ'}->get_queue_clist();
	$heap->{'DCC_CLIST'} = $heap->{'GTKOBJ'}->get_dcc_clist();
	$heap->{'NETWORK_CLIST'} = $heap->{'GTKOBJ'}->get_network_clist();
	$heap->{'CHANNEL_CLIST'} = $heap->{'GTKOBJ'}->get_channel_clist();
	$heap->{'STATUS_TEXT'} = $heap->{'GTKOBJ'}->get_status_text();
	$heap->{'STATUS_BAR'} = $heap->{'GTKOBJ'}->get_status_bar();
	$heap->{'SEARCH_ENTRY'} = $heap->{'GTKOBJ'}->get_search_entry();
	$heap->{'OFFER_POPUP'} = $heap->{'GTKOBJ'}->get_offer_popup();
	$heap->{'QUEUE_POPUP'} = $heap->{'GTKOBJ'}->get_queue_popup();
	$heap->{'DCC_POPUP'} = $heap->{'GTKOBJ'}->get_dcc_popup();
	$heap->{'NETWORK_POPUP'} = $heap->{'GTKOBJ'}->get_network_popup();
	$heap->{'CHANNEL_POPUP'} = $heap->{'GTKOBJ'}->get_channel_popup();
	$heap->{'FILE_MENU'} = $heap->{'GTKOBJ'}->get_file_menu();

	$heap->{'WINDOW'}->set_title("Buttler - Version " . VERSION);

	$heap->{'OFFER_CLIST'}->signal_connect("click-column", $session->postback("gtk_title_leftclick"));
	$heap->{'OFFER_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_leftclick"));
	$heap->{'OFFER_CLIST'}->signal_connect("button_press_event", $session->postback("gtk_row_rightclick"));
	$heap->{'OFFER_CLIST'}->signal_connect("select_row",$session->postback("gtk_row_select"));
	$heap->{'OFFER_CLIST'}->signal_connect("unselect_row",$session->postback("gtk_row_unselect"));
	$heap->{'OFFER_SEARCH_COLUMN'} = OFFER_CLIST_DEFAULT_SORT;
	$heap->{'OFFER_SEARCH_ASCENDING'} = TRUE;

	$heap->{'QUEUE_CLIST'}->signal_connect("click-column", $session->postback("gtk_title_leftclick"));
	$heap->{'QUEUE_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_leftclick"));
	$heap->{'QUEUE_CLIST'}->signal_connect("button_press_event", $session->postback("gtk_row_rightclick"));
	$heap->{'QUEUE_CLIST'}->signal_connect("select_row",$session->postback("gtk_row_select"));
	$heap->{'QUEUE_CLIST'}->signal_connect("unselect_row",$session->postback("gtk_row_unselect"));
	$heap->{'QUEUE_SEARCH_COLUMN'} = QUEUE_CLIST_DEFAULT_SORT;
	$heap->{'QUEUE_SEARCH_ASCENDING'} = TRUE;

	$heap->{'DCC_CLIST'}->signal_connect("click-column", $session->postback("gtk_title_leftclick"));
	$heap->{'DCC_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_leftclick"));
	$heap->{'DCC_CLIST'}->signal_connect("button_press_event", $session->postback("gtk_row_rightclick"));
	$heap->{'DCC_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_select"));
	$heap->{'DCC_CLIST'}->signal_connect("unselect_row", $session->postback("gtk_row_unselect"));
	$heap->{'DCC_SEARCH_COLUMN'} = DCC_CLIST_DEFAULT_SORT;
	$heap->{'DCC_SEARCH_ASCENDING'} = TRUE;

	$heap->{'NETWORK_CLIST'}->signal_connect("click-column", $session->postback("gtk_title_leftclick"));
	$heap->{'NETWORK_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_leftclick"));
	$heap->{'NETWORK_CLIST'}->signal_connect("button_press_event", $session->postback("gtk_row_rightclick"));
	$heap->{'NETWORK_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_select"));
	$heap->{'NETWORK_CLIST'}->signal_connect("unselect_row", $session->postback("gtk_row_unselect"));
	$heap->{'NETWORK_SEARCH_COLUMN'} = NETWORK_CLIST_DEFAULT_SORT;
	$heap->{'NETWORK_SEARCH_ASCENDING'} = TRUE;

	$heap->{'CHANNEL_CLIST'}->signal_connect("click-column", $session->postback("gtk_title_leftclick"));
	$heap->{'CHANNEL_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_leftclick"));
	$heap->{'CHANNEL_CLIST'}->signal_connect("button_press_event", $session->postback("gtk_row_rightclick"));
	$heap->{'CHANNEL_CLIST'}->signal_connect("select_row", $session->postback("gtk_row_select"));
	$heap->{'CHANNEL_CLIST'}->signal_connect("unselect_row", $session->postback("gtk_row_unselect"));
	$heap->{'CHANNEL_SEARCH_COLUMN'} = CHANNEL_CLIST_DEFAULT_SORT;
	$heap->{'CHANNEL_SEARCH_ASCENDING'} = TRUE;

	$heap->{'SEARCH_ENTRY'}->signal_connect("activate",$session->postback("gtk_search_new"));
	
	$heap->{'FILE_MENU'}->{'EXIT'}->signal_connect("activate",sub { exit });
	$heap->{'OFFER_POPUP'}->{'REQUEST'}->signal_connect("activate",$session->postback("gtk_offer_popup"));
	$heap->{'OFFER_POPUP'}->{'SELECT_ALL'}->signal_connect("activate",$session->postback("gtk_offer_popup"));
	$heap->{'OFFER_POPUP'}->{'UNSELECT_ALL'}->signal_connect("activate",$session->postback("gtk_offer_popup"));
	$heap->{'OFFER_POPUP'}->{'PURGE'}->signal_connect("activate",$session->postback("gtk_offer_popup"));

	$heap->{'QUEUE_POPUP'}->{'CANCEL'}->signal_connect("activate",$session->postback("gtk_queue_popup"));
	$heap->{'QUEUE_POPUP'}->{'SELECT_ALL'}->signal_connect("activate",$session->postback("gtk_queue_popup"));	
	$heap->{'QUEUE_POPUP'}->{'UNSELECT_ALL'}->signal_connect("activate",$session->postback("gtk_queue_popup"));	
	
	$heap->{'DCC_POPUP'}->{'STOP'}->signal_connect("activate",$session->postback("gtk_dcc_popup"));
	$heap->{'DCC_POPUP'}->{'STOP_DELETE'}->signal_connect("activate",$session->postback("gtk_dcc_popup"));
	$heap->{'DCC_POPUP'}->{'SELECT_ALL'}->signal_connect("activate",$session->postback("gtk_dcc_popup"));	
	$heap->{'DCC_POPUP'}->{'UNSELECT_ALL'}->signal_connect("activate",$session->postback("gtk_dcc_popup"));	
	
	$heap->{'NETWORK_POPUP'}->{'ENABLE'}->signal_connect("activate",$session->postback("gtk_network_popup"));
	$heap->{'CHANNEL_POPUP'}->{'JOIN'}->signal_connect("activate",$session->postback("gtk_channel_popup"));
	# stop buttler if window is closed
	$kernel->signal_ui_destroy($heap->{'WINDOW'});

	my ($min,$hour) = (localtime())[1,2];
        my $line = sprintf("%#.2d:%#.2d Buttler Version " . VERSION ." by " . AUTHOR_EMAIL . " Started.",$hour,$min);
        $heap->{'STATUS_TEXT'}->insert(undef,undef,undef,$line);


	# start refresh for dcc and statusbar
	$kernel->delay("gtk_dcc_refresh",1);
	$kernel->delay("gtk_statusbar_refresh",1);
	$kernel->delay("gtk_offer_refresh",GTK_OFFER_REFRESH);
	$kernel->delay("gtk_channel_refresh",1);
	$kernel->delay("gtk_network_refresh",1);
	$kernel->delay("gtk_queue_refresh",1);
}

# clicking on the title of the columns
sub gtk_title_leftclick {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];

	my $widget = shift @{$info};
	my $column = shift @{$info};


	if($widget == $heap->{'OFFER_CLIST'}) {
		if($heap->{'OFFER_SEARCH_COLUMN'} == $column) {
			$heap->{'OFFER_SEARCH_ASCENDING'} = 1 - $heap->{'OFFER_SEARCH_ASCENDING'};
		} else {
			$heap->{'OFFER_SEARCH_ASCENDING'} = TRUE;
		}
		$widget->set_sort_type(($heap->{'OFFER_SEARCH_ASCENDING'} ? "ascending" : "descending"));
		$heap->{'OFFER_SEARCH_COLUMN'} = $column;

	} elsif($widget == $heap->{'QUEUE_CLIST'}) {
		if($heap->{'QUEUE_SEARCH_COLUMN'} == $column) {
			$heap->{'QUEUE_SEARCH_ASCENDING'} = 1 - $heap->{'QUEUE_SEARCH_ASCENDING'};
		} else {
			$heap->{'QUEUE_SEARCH_ASCENDING'} = TRUE;
		}
		$widget->set_sort_type(($heap->{'QUEUE_SEARCH_ASCENDING'} ? "ascending" : "descending"));
		$heap->{'QUEUE_SEARCH_COLUMN'} = $column;

	} elsif($widget == $heap->{'DCC_CLIST'}) {
		if($heap->{'DCC_SEARCH_COLUMN'} == $column) {
			$heap->{'DCC_SEARCH_ASCENDING'} = 1 - $heap->{'DCC_SEARCH_ASCENDING'};
		} else {
			$heap->{'DCC_SEARCH_ASCENDING'} = TRUE;
		}
		$widget->set_sort_type(($heap->{'DCC_SEARCH_ASCENDING'} ? "ascending" : "descending"));
		$heap->{'DCC_SEARCH_COLUMN'} = $column;

	} elsif($widget == $heap->{'NETWORK_CLIST'}) {
		if($heap->{'NETWORK_SEARCH_COLUMN'} == $column) {
			$heap->{'NETWORK_SEARCH_ASCENDING'} = 1 - $heap->{'NETWORK_SEARCH_ASCENDING'};
		} else {
			$heap->{'NETWORK_SEARCH_ASCENDING'} = TRUE;
		}
		$widget->set_sort_type(($heap->{'NETWORK_SEARCH_ASCENDING'} ? "ascending" : "descending"));
		$heap->{'NETWORK_SEARCH_COLUMN'} = $column;

	} elsif($widget == $heap->{'CHANNEL_CLIST'}) {
		if($heap->{'CHANNEL_SEARCH_COLUMN'} == $column) {
			$heap->{'CHANNEL_SEARCH_ASCENDING'} = 1 - $heap->{'CHANNEL_SEARCH_ASCENDING'};
		} else {
			$heap->{'CHANNEL_SEARCH_ASCENDING'} = TRUE;
		}
		$widget->set_sort_type(($heap->{'CHANNEL_SEARCH_ASCENDING'} ? "ascending" : "descending"));
		$heap->{'CHANNEL_SEARCH_COLUMN'} = $column;

	}

	$widget->set_sort_column($column);
	$widget->sort();
	
}

# clicking on a row in the table
sub gtk_row_leftclick {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];

	my $widget = shift @{$info};
	my $row = shift @{$info};
	my $column = shift @{$info};
	my $event = shift @{$info};	
	
	if($widget == $heap->{'OFFER_CLIST'}) {
		# if double click event
		if($event->{'type'} eq "2button_press") {
			my $network = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_NETWORK);
			my $nick = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_BOT);
			my $pack = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_PACK);
			my $offer = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_DESCRIPTION);
			print "Adding ($pack) $offer from $nick\@$network to queue\n";
			$kernel->post(QUEUE_ALIAS,"queue_add_item",$network,$nick,$pack,$offer);
			$kernel->post(GTK_ALIAS,"gtk_offer_refresh");
		}
	} elsif($widget == $heap->{'QUEUE_CLIST'}) {
		# if double click event
		if($event->{'type'} eq "2button_press") {
			my $network = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_NETWORK);
			my $nick = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_BOT);
			my $pack = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_PACK);
			$kernel->post(QUEUE_ALIAS,"queue_remove_item",$network,$nick,$pack);
		}
	} elsif($widget == $heap->{'DCC_CLIST'}) {
		# if double click event
		if($event->{'type'} eq "2button_press") {
			my $dccid = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_ID);
			my $nick = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_BOT);
			my $file = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_FILE);
		
			if(defined($dcc->{'DCC'}->{$dccid})) {
				$kernel->post("$dcc->{'DCC'}->{$dccid}->{'NETWORK'}-bot" => dcc_close => $dccid);
				$kernel->post(GTK_ALIAS,"gtk_status","Transfer: Stopping dcc connection from $nick - $file");
				$dcc->{'DCC'}->{$dccid}->{'ACTIVE'} = FALSE;
			} else {
				print "DCC remove, unknown dcc id: $dccid\n";
			}
		}
	} elsif($widget == $heap->{'NETWORK_CLIST'}) {
		# if double click event
		if($event->{'type'} eq "2button_press") {
			my $network = $heap->{'NETWORK_CLIST'}->get_text($row,NETWORK_CLIST_NETWORK);
			if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISABLED) {
				$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISCONNECTED;
				$butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} = 0;
				$kernel->post("$network-session" => "network_connect");
				$kernel->post(GTK_ALIAS,"gtk_status","Network: $network re-enabled");
			}
		}
	} elsif($widget == $heap->{'CHANNEL_CLIST'}) {
		# if double click event
		if($event->{'type'} eq "2button_press") {
			my $network = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_NETWORK);
			my $channel = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_CHANNEL);
			if($butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} != CHANNEL_STATUS_SUCCESS && $butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} != CHANNEL_STATUS_JOIN) {
				$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_JOIN;
				$kernel->post("$network-bot" => join => $channel);
				$kernel->post(GTK_ALIAS,"gtk_status","Channel: Forcing join of $channel on $network network");
			}	
		}
	}
	
}

sub gtk_row_rightclick {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];

	my $widget = shift @{$info};
	my $event = shift @{$info};

	if($widget == $heap->{'OFFER_CLIST'}) {
		if($event->{'type'} eq "button_press" && $event->{'source'} eq "mouse" && $event->{'button'} eq 3) {
			$heap->{'OFFER_POPUP'}->{'MENU'}->popup(undef,undef,$event->{'time'},$event->{'button'},undef);
		}
	} elsif($widget == $heap->{'QUEUE_CLIST'}) {
		if($event->{'type'} eq "button_press" && $event->{'source'} eq "mouse" && $event->{'button'} eq 3) {	
			$heap->{'QUEUE_POPUP'}->{'MENU'}->popup(undef,undef,$event->{'time'},$event->{'button'},undef);
		}
	} elsif($widget == $heap->{'DCC_CLIST'}) {
		if($event->{'type'} eq "button_press" && $event->{'source'} eq "mouse" && $event->{'button'} eq 3) {	
			$heap->{'DCC_POPUP'}->{'MENU'}->popup(undef,undef,$event->{'time'},$event->{'button'},undef);
		}
	} elsif($widget == $heap->{'NETWORK_CLIST'}) {
		if($event->{'type'} eq "button_press" && $event->{'source'} eq "mouse" && $event->{'button'} eq 3) {	
			$heap->{'NETWORK_POPUP'}->{'MENU'}->popup(undef,undef,$event->{'time'},$event->{'button'},undef);
		}
	} elsif($widget == $heap->{'CHANNEL_CLIST'}) {
		if($event->{'type'} eq "button_press" && $event->{'source'} eq "mouse" && $event->{'button'} eq 3) {	
			$heap->{'CHANNEL_POPUP'}->{'MENU'}->popup(undef,undef,$event->{'time'},$event->{'button'},undef);
		}
	} 
}		

sub gtk_channel_popup {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];
	
	my $widget = shift @{$info};

	if($widget == $heap->{'CHANNEL_POPUP'}->{'JOIN'}) {
		for my $network (keys %{$heap->{'CHANNEL_SELECT'}}) {
			for my $channel (keys %{$heap->{'CHANNEL_SELECT'}->{$network}}) {
				if($butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} != CHANNEL_STATUS_SUCCESS && $butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} != CHANNEL_STATUS_JOIN) {
					$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} = CHANNEL_STATUS_JOIN;
					$kernel->post("$network-bot" => join => $channel);
					$kernel->post(GTK_ALIAS,"gtk_status","Channel: Forcing join of $channel on $network network");
				}
			}
		}
	} 
}

sub gtk_network_popup {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];
	
	my $widget = shift @{$info};

	if($widget == $heap->{'NETWORK_POPUP'}->{'ENABLE'}) {
		for my $network (keys %{$heap->{'NETWORK_SELECT'}}) {
			if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISABLED) {
				$butt->{'NETWORKS'}->{$network}->{'STATUS'} = NETWORK_STATUS_DISCONNECTED;
				$butt->{'NETWORKS'}->{$network}->{'ATTEMPTS'} = 0;
				$kernel->post("$network-session" => "network_connect");
				$kernel->post(GTK_ALIAS,"gtk_status","Network: $network re-enabled");
			}
		}

	} 
}

sub gtk_dcc_popup {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];
	
	my $widget = shift @{$info};

	if($widget == $heap->{'DCC_POPUP'}->{'STOP'} || $widget == $heap->{'DCC_POPUP'}->{'STOP_DELETE'}) {
		for my $dccid (keys %{$heap->{'DCC_SELECT'}}) {
			for my $nick (keys %{$heap->{'DCC_SELECT'}->{$dccid}}) {
				for my $file (keys %{$heap->{'DCC_SELECT'}->{$dccid}->{$nick}}) {
					if(defined($dcc->{'DCC'}->{$dccid})) {
						$kernel->post("$dcc->{'DCC'}->{$dccid}->{'NETWORK'}-bot" => dcc_close => $dccid);
						$kernel->post(GTK_ALIAS,"gtk_status","Transfer: Stopping dcc connection from $nick - $file");
						$dcc->{'DCC'}->{$dccid}->{'ACTIVE'} = FALSE;
						if($widget == $heap->{'DCC_POPUP'}->{'STOP_DELETE'}) {
							$dcc->{'DCC'}->{$dccid}->{'DELETE'} = TRUE;
						}
					} else {
						print "DCC remove, unknown dcc id: $dccid\n";
					}
				}
			}
		}

	} elsif($widget == $heap->{'DCC_POPUP'}->{'SELECT_ALL'}) {
		$heap->{'DCC_CLIST'}->select_all();
	} elsif($widget == $heap->{'DCC_POPUP'}->{'UNSELECT_ALL'}) {
		$heap->{'DCC_CLIST'}->unselct_all();	
	}
}


sub gtk_queue_popup {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];
	
	my $widget = shift @{$info};

	if($widget == $heap->{'QUEUE_POPUP'}->{'CANCEL'}) {
		for my $network (keys %{$heap->{'QUEUE_SELECT'}}) {
			for my $nick (keys %{$heap->{'QUEUE_SELECT'}->{$network}}) {
				for my $pack (keys %{$heap->{'QUEUE_SELECT'}->{$network}->{$nick}}) {
					$kernel->post(QUEUE_ALIAS,"queue_remove_item",$network,$nick,$pack);
				}
			}
		}
		delete($heap->{'QUEUE_SELECT'});
	} elsif($widget == $heap->{'QUEUE_POPUP'}->{'SELECT_ALL'}) {
		$heap->{'QUEUE_CLIST'}->select_all();
	} elsif($widget == $heap->{'QUEUE_POPUP'}->{'UNSELECT_ALL'}) {
		$heap->{'QUEUE_CLIST'}->unselect_all();
	}
}

sub gtk_offer_popup {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];
	
	my $widget = shift @{$info};

	if($widget == $heap->{'OFFER_POPUP'}->{'REQUEST'}) {
		for my $network (keys %{$heap->{'OFFER_SELECT'}}) {
			for my $nick (keys %{$heap->{'OFFER_SELECT'}->{$network}}) {
				for my $pack (keys %{$heap->{'OFFER_SELECT'}->{$network}->{$nick}}) {
					$kernel->post(QUEUE_ALIAS,"queue_add_item",$network,$nick,$pack,$heap->{'OFFER_SELECT'}->{$network}->{$nick}->{$pack});
					print "Adding ($pack) $heap->{'OFFER_SELECT'}->{$network}->{$nick}->{$pack} from $nick\@$network to queue\n";
				}
			}
		}
		delete($heap->{'OFFER_SELECT'});
		$heap->{'OFFER_CLIST'}->unselect_all();

	} elsif($widget == $heap->{'OFFER_POPUP'}->{'SELECT_ALL'}) {
		$heap->{'OFFER_CLIST'}->select_all();
	} elsif($widget == $heap->{'OFFER_POPUP'}->{'UNSELECT_ALL'}) {
		$heap->{'OFFER_CLIST'}->unselect_all();
	} elsif($widget == $heap->{'OFFER_POPUP'}->{'PURGE'}) {
		# clean up offers in $butt struct of old packs or disconnected users/networks
		$stats->{'TOTAL_PACKS'} = 0;
		$heap->{'OFFER_CLIST'}->freeze();
		$heap->{'OFFER_CLIST'}->clear();
		for my $network (keys %{$butt->{'NETWORKS'}}) {
			# if network is disconnected/disabled axe all offers from that network
			if($butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISCONNECTED || $butt->{'NETWORKS'}->{$network}->{'STATUS'} == NETWORK_STATUS_DISABLED) {
				delete($butt->{'NETWORKS'}->{$network}->{'USERS'});
			}
			for my $user (keys %{$butt->{'NETWORKS'}->{$network}->{'USERS'}}) {
				# if bot is offline or last_update > then timeout delete user
				if($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'STATUS'} == BOT_STATUS_OFFLINE || time - $butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'LAST_UPDATE'} >  GTK_OFFER_STALE_TIME) {
					delete($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user});
				} else {
					for my $pack (keys %{$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}}) {
						my $queue = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'OPEN_QUEUE'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'TOTAL_QUEUE'}";
						my $slots = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'OPEN_SLOTS'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'TOTAL_SLOTS'}";
						if($heap->{'SEARCH_REGEX'} eq "") {
							$heap->{'OFFER_CLIST'}->append($network,$user,$pack,$slots,$queue,$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'SIZE'},$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'OFFER'});	
						} elsif($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'OFFER'} =~ /$heap->{'SEARCH_REGEX'}/i) {
							$heap->{'OFFER_CLIST'}->append($network,$user,$pack,$slots,$queue,$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'SIZE'},$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'OFFER'});	
						}
						$stats->{'TOTAL_PACKS'}++;
					}
				}
			}
		}
		$heap->{'OFFER_CLIST'}->sort();
		$heap->{'OFFER_CLIST'}->thaw();
	}

	$kernel->post(GTK_ALIAS,"gtk_offer_refresh");
}

sub gtk_row_select {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];

	my $widget = shift @{$info};
	my $row = shift @{$info};

	if($widget == $heap->{'OFFER_CLIST'}) {
		my $network = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_NETWORK);
		my $nick = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_BOT);
		my $pack = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_PACK);
		my $offer = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_DESCRIPTION);
		$heap->{'OFFER_SELECT'}->{$network}->{$nick}->{$pack} = $offer;
	} elsif($widget == $heap->{'QUEUE_CLIST'}) {
		my $network = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_NETWORK);
		my $nick = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_BOT);
		my $pack = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_PACK);
		$heap->{'QUEUE_SELECT'}->{$network}->{$nick}->{$pack} = TRUE;
	} elsif($widget == $heap->{'DCC_CLIST'}) {
		my $dccid = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_ID);
		my $nick = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_BOT);
		my $file = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_FILE);
		$heap->{'DCC_SELECT'}->{$dccid}->{$nick}->{$file} = TRUE;
	} elsif($widget == $heap->{'NETWORK_CLIST'}) {
		my $network = $heap->{'NETWORK_CLIST'}->get_text($row,NETWORK_CLIST_NETWORK);
		$heap->{'NETWORK_SELECT'}->{$network} = TRUE;
	} elsif($widget == $heap->{'CHANNEL_CLIST'}) {
		my $network = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_NETWORK);	
		my $channel = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_CHANNEL);	
		$heap->{'CHANNEL_SELECT'}->{$network}->{$channel} = TRUE;
	} else {
		print "Unknown widget\n";
	}
}

sub gtk_row_unselect {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $info = @_[ARG1];

	my $widget = shift @{$info};
	my $row = shift @{$info};

	if($widget == $heap->{'OFFER_CLIST'}) {
		my $network = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_NETWORK);
		my $nick = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_BOT);
		my $pack = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_PACK);

		delete($heap->{'OFFER_SELECT'}->{$network}->{$nick}->{$pack});

		if(scalar(keys %{$heap->{'OFFER_SELECT'}->{$network}->{$nick}}) == 0) {
			delete($heap->{'OFFER_SELECT'}->{$network}->{$nick});
		}

		if(scalar(keys %{$heap->{'OFFER_SELECT'}->{$network}}) == 0) {
			delete($heap->{'OFFER_SELECT'}->{$network});
		}
	} elsif($widget == $heap->{'QUEUE_CLIST'}) {
		my $network = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_NETWORK);
		my $nick = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_BOT);
		my $pack = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_PACK);
		delete($heap->{'QUEUE_SELECT'}->{$network}->{$nick}->{$pack});

		if(scalar(keys %{$heap->{'QUEUE_SELECT'}->{$network}->{$nick}}) == 0) {
			delete($heap->{'QUEUE_SELECT'}->{$network}->{$nick});
		}

		if(scalar(keys %{$heap->{'QUEUE_SELECT'}->{$network}}) == 0) {
			delete($heap->{'QUEUE_SELECT'}->{$network});
		}

	} elsif($widget == $heap->{'DCC_CLIST'}) {
		my $dccid = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_ID);
		my $nick = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_BOT);
		my $file = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_FILE);
		delete($heap->{'DCC_SELECT'}->{$dccid}->{$nick}->{$file});

		if(scalar(keys %{$heap->{'DCC_SELECT'}->{$dccid}->{$nick}}) == 0) {
			delete($heap->{'DCC_SELECT'}->{$dccid}->{$nick});
		}
                
		if(scalar(keys %{$heap->{'DCC_SELECT'}->{$dccid}}) == 0) {
			delete($heap->{'DCC_SELECT'}->{$dccid});         
		}
	} elsif($widget == $heap->{'NETWORK_CLIST'}) {
		my $network = $heap->{'NETWORK_CLIST'}->get_text($row,NETWORK_CLIST_NETWORK);
		delete($heap->{'NETWORK_SELECT'}->{$network});
	} elsif($widget == $heap->{'CHANNEL_CLIST'}) {
		my $network = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_NETWORK);	
		my $channel = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_CHANNEL);	
		delete($heap->{'CHANNEL_SELECT'}->{$network}->{$channel});
	
		if(scalar(keys %{$heap->{'CHANNEL_SELECT'}->{$network}}) == 0) {
			delete($heap->{'CHANNEL_SELECT'}->{$network});
		}
	} else {
		print "Unknown widget\n";
	}

}


sub gtk_offer_refresh {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	
	my $color_network_offline = Gtk::Gdk::Color->parse_color(GTK_OFFER_COLOR_NETWORK_OFFLINE);
	my $color_bot_offline = Gtk::Gdk::Color->parse_color(GTK_OFFER_COLOR_BOT_OFFLINE);
	my $color_stale_offer = Gtk::Gdk::Color->parse_color(GTK_OFFER_COLOR_STALE_OFFER);
	my $color_queue = Gtk::Gdk::Color->parse_color(GTK_OFFER_COLOR_QUEUE);
	my $color_default = Gtk::Gdk::Color->parse_color(GTK_OFFER_COLOR_DEFAULT);	

	my $ctime = time;
	my $row = 0;

	$heap->{'OFFER_CLIST'}->freeze();

	while($heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_NETWORK) ne "") {
		my $network = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_NETWORK);
		my $nick = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_BOT);	
		my $pack = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_PACK);
		my $slots = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_SLOTS);
		my $queue = $heap->{'OFFER_CLIST'}->get_text($row,OFFER_CLIST_QUEUE);

		my $new_queue = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'}";
		my $new_slots = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'}";

		if($slots ne $new_slots) {
			$heap->{'OFFER_CLIST'}->set_text($row,OFFER_CLIST_SLOTS,$new_slots);
		}

		# update queue status
		if($queue ne $new_queue) {
			$heap->{'OFFER_CLIST'}->set_text($row,OFFER_CLIST_QUEUE,$new_queue);
		}


		# see if network is not connected.
		if($butt->{'NETWORKS'}->{$network}->{'STATUS'} != NETWORK_STATUS_CONNECTED) {
			$heap->{'OFFER_CLIST'}->set_background($row,$color_network_offline);
		
		# bot is offline
		} elsif($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} != BOT_STATUS_ONLINE) {
			$heap->{'OFFER_CLIST'}->set_background($row,$color_bot_offline);

		# no messages from bot for GTK_OFFER_STALE_OFFER
		} elsif(time - $butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'LAST_UPDATE'} > GTK_OFFER_STALE_TIME) {
			$heap->{'OFFER_CLIST'}->set_background($row,$color_stale_offer);

		} elsif($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'QUEUE'}){
			$heap->{'OFFER_CLIST'}->set_background($row,$color_queue);
		} else {
		 	$heap->{'OFFER_CLIST'}->set_background($row,$color_default);
		}
		$row++;
	}
	
	$heap->{'OFFER_CLIST'}->thaw();

	$kernel->delay("gtk_offer_refresh",GTK_OFFER_REFRESH);
}
# called when we need to redraw the queue clist
sub gtk_queue_refresh {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];

	my $queue = $kernel->call(QUEUE_ALIAS,"queue_get_queue");

	my $color_network_offline = Gtk::Gdk::Color->parse_color(GTK_QUEUE_COLOR_NETWORK_OFFLINE);
	my $color_bot_offline = Gtk::Gdk::Color->parse_color(GTK_QUEUE_COLOR_BOT_OFFLINE);
	my $color_stale_request = Gtk::Gdk::Color->parse_color(GTK_QUEUE_COLOR_STALE_REQUEST);
	my $color_default = Gtk::Gdk::Color->parse_color(GTK_QUEUE_COLOR_DEFAULT);	

	my $ctime = time;
	$heap->{'QUEUE_CLIST'}->freeze();
	my $vadjust = $heap->{'QUEUE_CLIST'}->get_vadjustment();
	my $value = $vadjust->value();

	$heap->{'QUEUE_CLIST'}->clear();

	if(defined($queue)) {
		my $row = 0;
		for my $network (sort keys %{$queue}) {
			for my $nick (sort keys %{$queue->{$network}}) {
				for my $item (sort @{$queue->{$network}->{$nick}}) {
					if($item->{'STATE'} != QUEUE_STATE_REMOVE) {
						$heap->{'QUEUE_CLIST'}->append($network,$nick,$item->{'PACK'},$item->{'INFO'},$item->{'OFFER'});
						if($butt->{'NETWORKS'}->{$network}->{'STATUS'} != NETWORK_STATUS_CONNECTED) {
							$heap->{'QUEUE_CLIST'}->set_background($row,$color_network_offline);
						} elsif($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'STATUS'} == BOT_STATUS_OFFLINE) {
							$heap->{'QUEUE_CLIST'}->set_background($row,$color_bot_offline);
						} elsif($item->{'STATE'} != QUEUE_STATE_TRANSFER && $item->{'STATE'} != QUEUE_STATE_LOCAL && $ctime - $item->{'TIME'} > GTK_QUEUE_STALE_TIME) {
							$heap->{'QUEUE_CLIST'}->set_background($row,$color_stale_request);
						} else {
							$heap->{'QUEUE_CLIST'}->set_background($row,$color_default);
						}

						$row++;
					}
				}
			}
		}
	}

	$vadjust->set_value($value);
	$heap->{'QUEUE_CLIST'}->sort();
	
	# relect rows.
	my $row = 0;
	my $total_rows = $heap->{'QUEUE_CLIST'}->rows();
	while($row < $total_rows) {
		my $network = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_NETWORK);
		my $nick = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_BOT);
		my $pack = $heap->{'QUEUE_CLIST'}->get_text($row,QUEUE_CLIST_PACK);
		if($heap->{'QUEUE_SELECT'}->{$network}->{$nick}->{$pack} == TRUE) {
			$heap->{'QUEUE_CLIST'}->select_row($row,0);
		}
		$row++;
	}

	$heap->{'QUEUE_CLIST'}->thaw();
	$kernel->delay("gtk_queue_refresh",1);
}

sub gtk_dcc_refresh {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];

	$heap->{'DCC_CLIST'}->freeze();
	my $vadjust = $heap->{'DCC_CLIST'}->get_vadjustment();
	my $value = $vadjust->value();

	$heap->{'DCC_CLIST'}->clear();

	$stats->{'SPEED'} = 0.0;
	for my $dccid (keys %{$dcc->{'DCC'}}) {
		if(!$dcc->{'DCC'}->{$dccid}->{'ACTIVE'}) {
			next;
		}

		my $timediff = time - $dcc->{'DCC'}->{$dccid}->{'START_TIME'};
		if($timediff <= 0) {
			$timediff = 1;
		}
		my $speed = sprintf("%.1f", $dcc->{'DCC'}->{$dccid}->{'BYTES_READ'} / ($timediff * 1024));
		$stats->{'SPEED'} = $stats->{'SPEED'} + $speed;

		$dcc->{'DCC'}->{$dccid}->{'FILE'} =~ /.*\/transferring\-(.*)$/;
		my $file = $1;
		my $percent = sprintf("%.1f%%", 100.0 * $dcc->{'DCC'}->{$dccid}->{'BYTES_READ'} / $dcc->{'DCC'}->{$dccid}->{'BYTES_TOTAL'});
		
		my $read = sprintf("%.1f", $dcc->{'DCC'}->{$dccid}->{'BYTES_READ'} / (1024 * 1024));
		my $total = sprintf("%.1fM", $dcc->{'DCC'}->{$dccid}->{'BYTES_TOTAL'} / (1024 * 1024));
	
		my $eta_string;
		if($speed > 0) {
			my $eta_seconds = ($dcc->{'DCC'}->{$dccid}->{'BYTES_TOTAL'} - $dcc->{'DCC'}->{$dccid}->{'BYTES_READ'}) / ($speed * 1024);
			$eta_string = strtimediff($eta_seconds);
		} else {
			$eta_string = "Never";
		}
		$heap->{'DCC_CLIST'}->append(
			$dccid,
			$dcc->{'DCC'}->{$dccid}->{'NICK'},
			$file,
			"$read/$total",
			$percent,
			$speed . "K/s",
			$eta_string);
	}

	$heap->{'DCC_CLIST'}->sort();
	$vadjust->set_value($value);
	
	# reselect rows

	my $row = 0;
	my $total_rows = $heap->{'DCC_CLIST'}->rows();

	while($row < $total_rows) {
		my $dccid = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_ID);
		my $nick = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_BOT);
		my $file = $heap->{'DCC_CLIST'}->get_text($row,DCC_CLIST_FILE);
                if($heap->{'DCC_SELECT'}->{$dccid}->{$nick}->{$file} == TRUE) {
			$heap->{'DCC_CLIST'}->select_row($row,0);
		}
		$row++;
	}

	$heap->{'DCC_CLIST'}->thaw();
	$kernel->delay("gtk_dcc_refresh",1);
}

sub gtk_network_refresh {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];

	$heap->{'NETWORK_CLIST'}->freeze();

	my $vadjust = $heap->{'NETWORK_CLIST'}->get_vadjustment();
	my $value = $vadjust->get_value();

	$heap->{'NETWORK_CLIST'}->clear();

	for my $network (keys %{$butt->{'NETWORKS'}}) {
		my $status = $butt->{'NETWORKS'}->{$network}->{'STATUS'};
		my $status_string;
		my $channels_in = 0;
		if($status == NETWORK_STATUS_DISABLED) {
			$status_string = "Network is disabled";
		} elsif($status == NETWORK_STATUS_CONNECTING) {
			$status_string = "Connecting to Network";
		} elsif($status == NETWORK_STATUS_DISCONNECTED) {
			$status_string = "Disconnected, reconnect soon";
		} elsif($status == NETWORK_STATUS_CONNECTED) {	
			$status_string = "Connected";
			for my $channel (keys %{$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}}) {
				if($butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'} == CHANNEL_STATUS_SUCCESS) {
					$channels_in++;
				}
			}
		} else {
			$status_string = "Invalid State";
		}

		# $butt wont have info on total channels
		my $channels_total = scalar(keys %{$config->{'NETWORKS'}->{$network}->{'CHANNELS'}});
		$heap->{'NETWORK_CLIST'}->append($network,$status_string,"$channels_in/$channels_total",$butt->{'NETWORKS'}->{$network}->{'INFO'});
	}

	
	$heap->{'NETWORK_CLIST'}->sort();

	# reselect rows that were previously selected.
	my $row = 0;
	my $total_rows = $heap->{'NETWORK_CLIST'}->rows();

	while($row < $total_rows) {
		my $network = $heap->{'NETWORK_CLIST'}->get_text($row,NETWORK_CLIST_NETWORK);
		if($heap->{'NETWORK_SELECT'}->{$network} == TRUE) {
			$heap->{'NETWORK_CLIST'}->select_row($row,0);
		}
		$row++;
	}

	$vadjust->set_value($value);
	$vadjust->changed();
	$heap->{'NETWORK_CLIST'}->thaw();

	$kernel->delay("gtk_network_refresh",1);
}

sub gtk_channel_refresh {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];

	$heap->{'CHANNEL_CLIST'}->freeze();
	my $vadjust = $heap->{'CHANNEL_CLIST'}->get_vadjustment();
	my $value = $vadjust->get_value();

	$heap->{'CHANNEL_CLIST'}->clear();

	for my $network (keys %{$butt->{'NETWORKS'}}) {
		for my $channel (keys %{$butt->{'NETWORKS'}->{$network}->{'CHANNELS'}}) {
			my $status = $butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'STATUS'};
			my $status_string;
			my $topic = "n/a";
			if($status == CHANNEL_STATUS_SUCCESS) {
				$status_string = "Success";
				$topic = $butt->{'NETWORKS'}->{$network}->{'CHANNELS'}->{$channel}->{'TOPIC'};
			} elsif($status == CHANNEL_STATUS_JOIN) {
				$status_string = "Joining";
			} elsif($status == CHANNEL_STATUS_BANNED) {
				$status_string = "Banned";
			} elsif($status == CHANNEL_STATUS_INVITE) {
				$status_string = "Invite only";
			} elsif($status == CHANNEL_STATUS_TOOMANY) {
				$status_string = "Too many channels";
			} elsif($status == CHANNEL_STATUS_FULL) {
				$status_string = "full";
			} elsif($status == CHANNEL_STATUS_ERROR) {
				$status_string = "Unknown Error";
			} elsif($status == CHANNEL_STATUS_KICKED) {
				$status_string = "Kicked";
			}
			$heap->{'CHANNEL_CLIST'}->append($network,$channel,$status_string,$topic);
		}
	}

	$heap->{'CHANNEL_CLIST'}->sort();


	# re-select rows
	my $row = 0;
	my $total_rows = $heap->{'CHANNEL_CLIST'}->rows();
	while($row < $total_rows) {
		my $network = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_NETWORK);	
		my $channel = $heap->{'CHANNEL_CLIST'}->get_text($row,CHANNEL_CLIST_CHANNEL);	
		if($heap->{'CHANNEL_SELECT'}->{$network}->{$channel} == TRUE) {
			$heap->{'CHANNEL_CLIST'}->select_row($row,0);
		}
		$row++;
	}

	$vadjust->set_value($value);
	$heap->{'CHANNEL_CLIST'}->thaw();
	$kernel->delay("gtk_channel_refresh",1);
}

sub gtk_statusbar_refresh {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];

	if(!defined($heap->{'CONTEXT_ID'})) {
		$heap->{'CONTEXT_ID'} = $heap->{'STATUS_BAR'}->get_context_id("StatusBar");
	}

	if(defined($heap->{'MESSAGE_ID'})) {
		$heap->{'STATUS_BAR'}->remove($heap->{'CONTEXT_ID'},$heap->{'MESSAGE_ID'});
	}

	$heap->{'STATUS_BAR'}->push($heap->{'CONTEXT_ID'}, " Stats: " .
		$stats->{'TOTAL_PACKS'} . " Offers in " . 
		$stats->{'TOTAL_CHANNELS'} . " Channels on " . 
		$stats->{'TOTAL_NETWORKS'} . " Networks. [" .
		scalar(keys %{$dcc->{'DCC'}}) . " Transfers \@ " .
		sprintf("%.1f",$stats->{'SPEED'}) . "K/s] - [Downloaded: " .
		strsize($stats->{'TOTAL_DOWNLOAD'}) . "]");
	
	$kernel->delay("gtk_statusbar_refresh",1);

}

# adds the passed line into the status text widget
sub gtk_status {
	my $heap = @_[HEAP];
	my ($line) = @_[ARG0];

	my ($min,$hour) = (localtime())[1,2];
	my $line = sprintf("\n%#.2d:%#.2d $line",$hour,$min);

	$heap->{'STATUS_TEXT'}->insert(undef,undef,undef,$line);
}

#
sub gtk_search_new {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];

	my $new_regex = $heap->{'SEARCH_ENTRY'}->get_text();

	if($heap->{'SEARCH_REGEX'} eq $new_regex) {
		$debug && print trace(), "Ignoring new search, same regex as before '$new_regex'\n";
		return;
	}

	$heap->{'SEARCH_REGEX'} = quotemeta($new_regex);
	$heap->{'OFFER_CLIST'}->freeze();
	$heap->{'OFFER_CLIST'}->clear();
	
	# remove any previously selected items
	delete($heap->{'OFFER_SELECT'});
	for my $network (keys %{$butt->{'NETWORKS'}}) {
		for my $user (keys %{$butt->{'NETWORKS'}->{$network}->{'USERS'}}) {
			for my $pack (keys %{$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}}) {
				my $queue = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'OPEN_QUEUE'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'TOTAL_QUEUE'}";
				my $slots = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'OPEN_SLOTS'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'TOTAL_SLOTS'}";
				if($heap->{'SEARCH_REGEX'} eq "") {
					$heap->{'OFFER_CLIST'}->append($network,$user,$pack,$slots,$queue,$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'SIZE'},$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'OFFER'});	
				} elsif($butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'OFFER'} =~ /$heap->{'SEARCH_REGEX'}/i) {
					$heap->{'OFFER_CLIST'}->append($network,$user,$pack,$slots,$queue,$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'SIZE'},$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$user}->{'PACKS'}->{$pack}->{'OFFER'});	
		
				}
			}
		}
	}

	$heap->{'OFFER_CLIST'}->sort();
	$heap->{'OFFER_CLIST'}->thaw();

	# force an refresh to get colors
	$kernel->yield("gtk_offer_refresh");
}

sub gtk_offer_row_add {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $network = @_[ARG0];
	my $nick = @_[ARG1];
	my $pack = @_[ARG2];
	my $size = @_[ARG3];
	my $description = @_[ARG4];

	if($heap->{'SEARCH_REGEX'} eq "" || $description =~ /$heap->{'SEARCH_REGEX'}/i) {
		my $queue = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_QUEUE'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_QUEUE'}";
		my $slots = "$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'OPEN_SLOTS'}/$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'TOTAL_SLOTS'}";
		$heap->{'OFFER_CLIST'}->append($network,$nick,$pack,$slots,$queue,$size,$description);
		$heap->{'OFFER_CLIST'}->sort();
	} 
}
##
## queue stuff
##

# called when the queue session is first created.  initialize stuff and start queue
sub queue_start {
	my $kernel = @_[KERNEL];

	$kernel->alias_set(QUEUE_ALIAS);

	$kernel->yield('queue_run');
}

# called every QUEUE_INTERVAL_TIME seconds to process the queue
sub queue_run {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	
	
	my $needupdate = FALSE;

	for my $network (keys %{$heap->{'QUEUE'}}) {
		for my $nick (keys %{$heap->{'QUEUE'}->{$network}}) {
			my $element = 0;
			
			# see if we need to remove queue item
			if(defined($heap->{'QUEUE'}->{$network}->{$nick}[0])) {
				if($heap->{'QUEUE'}->{$network}->{$nick}[0]->{'STATE'} == QUEUE_STATE_REMOVE) {
					$debug && print trace(), "Removing item from queue ($nick\@$network ($heap->{'QUEUE'}->{$network}->{$nick}[0]->{'PACK'})\n";
					$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$heap->{'QUEUE'}->{$network}->{$nick}[0]->{'PACK'}}->{'QUEUE'} = FALSE;
					shift @{$heap->{'QUEUE'}->{$network}->{$nick}};
					$needupdate = TRUE;
				}
			}

			for my $item (@{$heap->{'QUEUE'}->{$network}->{$nick}}) {

				# only element 0 in the array should be active
				if($element == 0) {
					if($item->{'STATE'} == QUEUE_STATE_NEW || $item->{'STATE'} == QUEUE_STATE_LOCAL) {
						# send request to bot
						$kernel->post(GTK_ALIAS,"gtk_status","Sent Request: $item->{'OFFER'} from $nick");
						if($config->{'NETWORKS'}->{$network}->{'DCCALLOW'}) {
							print "$network: SENDING DCCALLOW +$nick\n";
							$kernel->post("$network-bot" => sl => "DCCALLOW +$nick");
						}
						$kernel->post("$network-bot" => privmsg => $nick => "xdcc send $item->{'PACK'}");
						$debug && print trace(), "Sending request to $nick\@$network (xdcc send $item->{'PACK'}) for $item->{'OFFER'}\n";
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'STATE'} = QUEUE_STATE_REQUEST;
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'TIME'} = time();
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'INFO'} = "Sent request to bot";
						$needupdate = TRUE;
					} elsif($item->{'STATE'} == QUEUE_STATE_FULL || $item->{'STATE'} == QUEUE_STATE_OFFLINE || $item->{'STATE'} == QUEUE_STATE_TIMEOUT) { 
						# try to request the pack again after QUEUE_RETRY_INTERVAL
						if(time - $item->{'TIME'} > QUEUE_RETRY_INTERVAL) {
							$kernel->post(GTK_ALIAS,"gtk_status","Sent Request: $item->{'OFFER'} from $nick");
							if($config->{'NETWORKS'}->{$network}->{'DCCALLOW'}) {
                                                        	print "$network: SENDING DCCALLOW +$nick\n";
                                                        	$kernel->post("$network-bot" => sl => "DCCALLOW +$nick");
                                                	}
							$kernel->post("$network-bot" => privmsg => $nick => "xdcc send $item->{'PACK'}");
							$debug && print trace(), "Re-requesting pack to $nick\@$network (xdcc send $item->{'PACK'}) for $item->{'OFFER'}\n";
							$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'STATE'} = QUEUE_STATE_REQUEST;
							$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'TIME'} = time();
							$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'INFO'} = "Re-Sent request to bot";
							$needupdate = TRUE;
						}
					} elsif($item->{'STATE'} == QUEUE_STATE_REQUEST) {
						# bot didnt respond to our request
						if(time - $item->{'TIME'} > QUEUE_REQUEST_TIMEOUT) {
							$debug && print trace(), "$nick\@$network didnt respond to our request\n";
							$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'STATE'} = QUEUE_STATE_TIMEOUT;
							$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'TIME'} = time();
							$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'INFO'} = "Bot didnt respond to request";
							$needupdate = TRUE;
						}
					} elsif($item->{'STATE'} == QUEUE_STATE_ERROR) {
						# remove the item after some time
						if(time - $item->{'TIME'} > QUEUE_ERROR_TIMEOUT) {
							$debug && print trace(), "removing $nick\@$network error from queue\n";
							$kernel->post(GTK_ALIAS,"gtk_status","Queue: Error ($item->{'INFO'}) Removing $item->{'OFFER'} from $nick");
							$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'STATE'} = QUEUE_STATE_REMOVE;
							$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$item->{'PACK'}}->{'QUEUE'} = FALSE;
							$needupdate = TRUE;
						}
					} 

				} else {
					
					# error on first item, mark them all as errors
					if($heap->{'QUEUE'}->{$network}->{$nick}[0]->{'STATE'} == QUEUE_STATE_ERROR) {
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'STATE'} = QUEUE_STATE_ERROR;
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'TIME'} = time();
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'INFO'} = "error'd from previous queue item";
						$needupdate = TRUE;
					# new item, queue locally
					} elsif($item->{'STATE'} == QUEUE_STATE_NEW) {
						$kernel->post(GTK_ALIAS,"gtk_status","QUEUE: $item->{'OFFER'} from $nick queue'd locally.");
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'STATE'} = QUEUE_STATE_LOCAL;
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'TIME'} = time();
						$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'INFO'} = "Item Queue'd Locally";
						$debug && print trace(), "$nick\@$network queue item queue'd locally.\n";
						$needupdate = TRUE;
					} elsif($item->{'STATE'} == QUEUE_STATE_REMOVE) {
						$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$item->{'PACK'}}->{'QUEUE'} = FALSE;
					}

				}
		
				$element++;
			}

			if(scalar(@{$heap->{'QUEUE'}->{$network}->{$nick}}) == 0) {
				delete($heap->{'QUEUE'}->{$network}->{$nick});
			}
			
		}

		if(scalar(keys %{$heap->{'QUEUE'}->{$network}}) == 0) {
			delete($heap->{'QUEUE'}->{$network});
		}
	}

	if($needupdate) {
		$kernel->post(GTK_ALIAS,"gtk_offer_refresh");
	}
	$kernel->delay('queue_run',QUEUE_RUN_INTERVAL);
}

# called when a new item is to be put into the queue
sub queue_add_item {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $network = @_[ARG0];
	my $nick = @_[ARG1];
	my $pack = @_[ARG2];
	my $offer = @_[ARG3];
	
	$nick =~ tr/[A-Z]/[a-z]/;

	## check if item already exists
	my $newitem = TRUE;
	if(defined($heap->{'QUEUE'}->{$network}->{$nick})) {
		for my $item (@{$heap->{'QUEUE'}->{$network}->{$nick}}) {
			if($item->{'PACK'} eq $pack) {
				$newitem = FALSE;
			}
		}
	}

	if($newitem) {
		my $item;
		$item->{'PACK'} = $pack;
		$item->{'OFFER'} = $offer;
		$item->{'STATE'} = QUEUE_STATE_NEW;
		$item->{'INFO'} = "New Queue Item";
		$butt->{'NETWORKS'}->{$network}->{'USERS'}->{$nick}->{'PACKS'}->{$pack}->{'QUEUE'} = TRUE;
		push @{$heap->{'QUEUE'}->{$network}->{$nick}}, $item;
		$debug && print trace(), "add Item to queue $nick\@$network $pack $offer.\n";
	} else {
		$debug && print trace(), "Item already in queue.\n";
	}
}

# called via $kernel->call() returns if passed network/user is in the queue
sub queue_valid_user {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $network = @_[ARG0];
	my $nick = @_[ARG1];

	if(defined($heap->{'QUEUE'}->{$network}->{$nick})) {
		return TRUE;
	} 
		
	return FALSE;
}

# called to remove a queue item
sub queue_remove_item {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $network = @_[ARG0];
	my $nick = @_[ARG1];
	my $pack = @_[ARG2];

	# try and find item
	my $founditem = FALSE;
	my $element = 0;
	$nick =~ tr/[A-Z]/[a-z]/;

	my $remove_item;
	if(defined($heap->{'QUEUE'}->{$network}->{$nick})) {
		for my $item (@{$heap->{'QUEUE'}->{$network}->{$nick}}) {
			if($item->{'PACK'} eq $pack) {

				# remotly queue'd let bot know we are removing
				if($item->{'STATE'} == QUEUE_STATE_REMOTE || $item->{'STATE'} == QUEUE_STATE_REQUEST) {
					$kernel->post("$network-bot" => privmsg => $nick => "xdcc send remove");
				}

				## if transfer need to stop the dcc transfer
				$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'STATE'} = QUEUE_STATE_REMOVE;
				$heap->{'QUEUE'}->{$network}->{$nick}[$element]->{'INFO'} = "Item removed from queue";
				$remove_item = $item;
				$founditem = TRUE;
			}
			$element++;
		}
	}

	if($founditem) {
		$kernel->post(GTK_ALIAS,"gtk_status","Queue: Removing $remove_item->{'OFFER'} from $nick");
		$debug && print trace(), "Item removed from queue\n";
	} else {
		$debug && print trace(), "Item not found for removal\n";
	}
}

# called to clear the entire queue
sub queue_clear {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];

	$debug && print trace(), "clearing queue\n";
	delete($heap->{'QUEUE'});
}

# $kernel->call() returns the current queue
sub queue_get_queue {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	return($heap->{'QUEUE'});
}

# called to dump the contents of the queue to stdout
sub queue_dump {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];


	print "Queue DUMP\n---------------------------\n";
	if(defined($heap->{'QUEUE'})) {
		for my $network (keys %{$heap->{'QUEUE'}}) {
			print "  $network\n";
			for my $nick (keys %{$heap->{'QUEUE'}->{$network}}) {
				print "    $nick\n";
				for my $item (@{$heap->{'QUEUE'}->{$network}->{$nick}}) {
					print "      ($item->{'PACK'}) $item->{'OFFER'} : $item->{'STATE'} : $item->{'INFO'}\n";
				}
			}
		}
	} else {
		print "  Queue Empty\n";
	}
	print "---------------------------\n";
}

# incoming dcc connection, verify its valid and then accept it
sub queue_incoming_dcc {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $dccid = @_[ARG0];
	my $nick = @_[ARG1];
	my $network = @_[ARG2];
	my $file = @_[ARG3];

	$nick =~ tr/[A-Z]/[a-z]/;

	# assume its a valid dcc request if this is defined
	if(defined($heap->{'QUEUE'}->{$network}->{$nick})) {
		$kernel->post("$network-bot" => dcc_accept => $dccid => $file);
	} else {
		$file =~ /\/transferring\-(.*)/;
		$kernel->post(GTK_ALIAS,"gtk_status","DCC: Ignoring request from $nick\@$network - $1");
	}
}

# called to update the status of a queue item
# you can only update the first element of a bots queue
sub queue_update_item {
	my $kernel = @_[KERNEL];
	my $heap = @_[HEAP];
	my $network = @_[ARG0];
	my $nick = @_[ARG1];
	my $state = @_[ARG2];
	my $info = @_[ARG3];

	my $founditem = FALSE;

	$nick =~ tr/[A-Z]/[a-z]/;

	if(defined($heap->{'QUEUE'}->{$network}->{$nick})) {
		# dont allow updates to items set to be removed 
		if($heap->{'QUEUE'}->{$network}->{$nick}[0]->{'STATE'} != QUEUE_STATE_REMOVE) {
			$heap->{'QUEUE'}->{$network}->{$nick}[0]->{'STATE'} = $state;
			$heap->{'QUEUE'}->{$network}->{$nick}[0]->{'INFO'} = $info;
			$heap->{'QUEUE'}->{$network}->{$nick}[0]->{'TIME'} = time;
			$founditem = TRUE;
		}
	}

	if($founditem) {
		$debug && print trace(), "Item updated\n";
	} else {
		$debug && print trace(), "Item not found for update\n";
	}

}


##
## misc functions
##

# needs to be filled in
sub random_string {
	my ($min,$max) = @_;
	my $nick;

	 my @text_array = ("a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z","0","1","2","3","4","5","6","7","8","9");

	my $length = $min + sprintf("%d",rand($max - $min + 1));
	if($length > $max) {   
		$length = $max;
	}

	my $count = 0;
	while($count <= $length) {
		$nick = $nick . random_element(@text_array);
		$count++;
	}
	return $nick;
}

# returns a random element from the passed array
sub random_element {
	return $_[rand scalar @_];
}

sub strsize {
	my ($size) = @_;

	if($size < 0) {
		$size = 0;
	}

	my $str;
	if($size > 1500000000) {
		$str = sprintf("%.2f Gigs", $size / (1024 * 1024 * 1024));
	} elsif($size > 1500000) {
		$str = sprintf("%.1f Megs", $size / (1024 * 1024));
	} elsif($size > 0) {
		$str = sprintf("%d KiloBytes", $size / 1024);
	} else {
		$str = "0 Bytes";
	}

	return $str;
}

sub strtimediff {
        my ($timediff) = @_;

        if($timediff < 0) {
                $timediff = 0;
        }
        my $str;
        if($timediff > 86400) {
                $str = sprintf("%.1f days", $timediff / 86400);
        } elsif($timediff > 5400) {
                $str = sprintf("%.1f hours", $timediff / 3600);
        } elsif($timediff > 90) {
                $str = sprintf("%.1f minutes", $timediff / 60);
        } else {
                $str = sprintf("%d seconds", $timediff);
        }

        return $str;

}

sub usage {
	print "$0 -c <dir> [-d] [-u <file>] [-m <file>]\n";
	print "Usage:\n";
	print "  -c <dir>	: where config dir is at.\n";
	print "  -d		: enable debugging output to stdout.\n";
	print "  -u <file>	: log all unparsed messages to <file>.\n";
	print "  -m <file>	: log messages from people in queue to <file>.\n";
}


